! -*-f90-*-
!
! Author:    jbakker
! Generated: Wed Oct  2 15:01:52 CEST 2002
! Usage:     transforms pixels coordinates in image to celestial coordinates
!
! Example:
!    use dal
!    use omsourceparameter_mod
!    use omsourcecelestial_mod
!    implicit none
!
!    mySet=dataSet(imageFileName, read)
!    call construct(myCelestial,mySet)
!
!    ! get the computed (RA,Dec) of image centre
!    call getCelestialCentre(myCelestial,ra,dec)
!
!    ! compute the (RA,Dec) of a source from its pixel coordinates
!    mySource%xCoord=50
!    mySource%yCoord=60
!    call setCelestial(myCelestial,mySource)
!
!    call destruct(myCelestial)
module omsourcecelestial_mod

   use cal
   use dal
   use sasdal
   use caloaldefs

   use errorHandling
   use sla
   use readModes_mod
!   use omsourceutils_mod
   use omconstants_mod
   use omsourceparameter_mod
   use omsourceparam_mod
  

   implicit none

   private   ! all, but
!   public :: OmCelestial
   public :: construct_celestial
   public :: destruct_celestial
   public :: getCelestialCentre
   public :: setCelestial

!   type OmCelestial
!      private
!      integer(int32) :: x0,y0   ! offset in image pixels
!      integer(int32) :: bx,by   ! binning in image pixels
!      real(double)   :: plateScale       
!      real(double)   :: rollAngle       
!      real(double)   :: ra,dec  ! of centre point

      ! info as above, but shortcuts, that are heavily used
!      real(double)   :: d_x0,d_y0
!      real(double)   :: d_bx,d_by
!      real(single)   :: sinRoll,cosRoll ! of rollangle
!   end type OmCelestial



!   interface construct
!      module procedure construct_
!   end interface

   
!   interface destruct
!      module procedure destruct_
!   end interface


   interface setCelestial
      module procedure setCelestial_,setCelestial_v
   end interface


    ! distortion offsets, needed as pointers, because the
    ! CAL expects it
    logical              :: isModuleInitialized=.false.
    real(single), pointer :: deltaX(:)
    real(single), pointer :: deltaY(:)

    real(double), private :: cdelt1     ! RA plate-scale (degrees and negative)
    real(double), private :: cdelt2     ! Dec plate-scale (degrees)
    real(double), private :: crpix1, crpix2
    real(double), private :: crval1, crval2
    logical, private :: isSkyImage = .false.    
  
contains
!************************************************************
! construct_
!************************************************************
!subroutine construct_(t, set)
subroutine construct_celestial(t, set)

    type(OmCelestial),       intent(out)          :: t
    type(DataSetT),          intent(in)           :: set !

    integer                 :: status ! allocation
    character(*), parameter :: routine="omsourcecelestial_mod.construct"

    ! --- initialise basics from set:
    call InitialiseCelestial(set, t%bx, t%by, t%ra, t%dec,&
           t%rollAngle, t%plateScale, t%x0, t%y0)
      
    ! --- now populate the helpers:
    t%d_x0  =dble(t%x0)
    t%d_y0  =dble(t%y0)
    t%d_bx  =dble(t%bx)
    t%d_by  =dble(t%by)
    t%sinRoll = dsin(t%rollAngle)
    t%cosRoll = dcos(t%rollAngle)

    ! --- and globals in this module:
    if (.not.isModuleInitialized) then
        allocate(deltaX(0 : 0),stat=status)
!        call assertOn(status/=0,routine//".deltaX")

        allocate(deltaY(0:0),stat=status)
!        call assertOn(status/=0,routine//".deltaY")
        isModuleInitialized=.true.
    end if
end subroutine construct_celestial



!subroutine destruct_(t)
subroutine destruct_celestial(t)

    type(OmCelestial),       intent(inout)          :: t

    t%ra=0.
    t%dec=0.
end subroutine destruct_celestial



   ! returns the (RA,Dec) of the image centre
subroutine getCelestialCentre(t,ra,dec)
    type(OmCelestial),       intent(in)     :: t       ! info object
    real(double),            intent(out)    :: ra,dec  ! of centre point
      
    ra =t%ra *radiansToDegrees
    dec=t%dec*radiansToDegrees
end subroutine getCelestialCentre


!**************************************************************
! subroutune setCelestial
!**************************************************************
subroutine setCelestial_(t, source)
      type(OmCelestial),       intent(in)    :: t       ! info object
      type(SourceParameters), intent(inout) :: source

      ! locals
     

      real(double)   :: xposcorr, yposcorr
      real(double)   :: xrad, yrad, rotxrad, rotyrad ! position in radians
      real(double)   :: xi_source
      real(double)   :: eta_source 
      real           :: rawx(0:0), rawy(0:0)  ! CAL input coords



    if(isskyImage) then
        call computeRaDec( source ) 


        return
    end if






      ! compute absolute position of object within OM frame
      xposcorr = dble((source%xPos - 0.5)) * t%d_bx + t%d_x0
      yposcorr = dble((source%yPos - 0.5)) * t%d_by + t%d_y0
      
      ! correct positions for pincushion/barrel distortion
      rawx(0) = real(xposcorr)
      rawy(0) = real(yposcorr)
      call CAL_omDistortion(rawx, rawy, deltaX, deltaY)
      xposcorr = xposcorr - dble(deltaX(0))    
      yposcorr = yposcorr - dble(deltaY(0))

      ! get offset relative to tangent point (at OM centre)
      xrad = xposcorr - xCentre 
      yrad = yposcorr - yCentre

      !  rotate x,y frame to tangent sky (RA, DEC) frame
      rotxrad = t%plateScale * (xrad * t%cosRoll - yrad * t%sinRoll)
      rotyrad = t%plateScale * (xrad * t%sinRoll + yrad * t%cosRoll)

      ! store standard coords for use astrometry correction
      xi_source  = -rotxrad   ! Tangent-plane coords
      eta_source =  rotyrad 

      ! Transform tangent-plane coordinates (xi_source, yi_source) 
      ! into spherical (sky) coordinates (RA, dec). ra_point,
      ! dec_point is the spherical coordinate of the tangent-point
      ! Also calculate the galactic longitude and latitude of
      ! each source
      call sla_dtp2s(xi_source, eta_source, &
           t%ra, t%dec, source%rightAscension, source%declination)

      ! return in degrees!
      source%rightAscension = source%rightAscension * radiansToDegrees
      source%declination   = source%declination * radiansToDegrees
end subroutine setCelestial_
!******************************************************
! subroutine setCelestial
!******************************************************
subroutine setCelestial_v(t, sources)
      type(OmCelestial),       intent(in)    :: t       ! info object
      type(SourceParameters), intent(inout) :: sources(:)

      integer :: i

      do i = 1, size(sources)
         call setCelestial(t, sources(i))
      end do
end subroutine setCelestial_v


!************************************************************
! Subroutine Initialise
!***********************************************************
subroutine InitialiseCelestial(aSet, xBinning, yBinning, & 
        ra_point, dec_point, &
        rollAngle, plateScale, x0, y0)            

      type(DataSetT), intent(in)  :: aSet
      integer(int32), intent(out) :: xBinning
      integer(int32), intent(out) :: yBinning
      real(double),   intent(out) :: dec_point  ! position OM tangent point
      real(double),   intent(out) :: ra_point   ! position OM tangent point
      real(double),   intent(out) :: rollangle  ! the roll angle
      real(double),   intent(out) :: plateScale ! the plate-scale
      integer(int32), intent(out) :: x0, y0     ! offset of the image


      type (SpacecraftAttitudeType) attitude
      character(10)           :: datamode   ! datamode attribute
      character(6)            :: omFilter   ! the OM filter

      integer             :: xbin, ybin         ! binning factors
      integer             :: xsize, ysize       ! size of the image

      integer                     :: OM_MODE    ! the om mode


      real(double)  :: rotationCorrection = 90.0_double * degreesToRadians ! Correction for coordinate error



      real(double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     


      real(double) :: time                        
      real(double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
      real(double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

      real(double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
      real(double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
      real(double) :: pointvec(3)        ! s/c pointing
      real(double) :: pointvec2(3)       


      real(double) :: euler_axis(3)              ! euler axis for rotation
      real(double), pointer :: boremat(:,:)       ! boresight offset matrix
      real(double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
      real(double) :: omvec(3)                   ! om pointing vector
      real(double) :: rotmat(3, 3)               ! rotation matrix


      real(double) :: r1                         ! modulus


      real(double)           :: lambda           ! wavelength for CAL
      logical(Bool) :: binBPE                    ! BPE binning on?


      integer(int32)           :: filterIdInt

      xbin = int32Attribute(aSet, "BINAX1")      ! binning factors
      ybin = int32Attribute(aSet, "BINAX2")

      if (xbin .ne. ybin) then
         call fatal("AsymBinning", "This version cannot deal with different binning factors in x and y")
      end if

      binBPE = booleanAttribute(aSet, "BINBPE")

      ! Calculate the image binning.
      xBinning = 2 ** xbin
      yBinning = 2 ** ybin

      if (binBPE) then
         xBinning = xBinning * 2
         yBinning = yBinning * 2
      end if

      !*******************************************************
      ! get the astrometry info keywords from the source list
      !*******************************************************
      ! RA, DEc & PA in radians...

      if( hasAttribute(aSet, "RA_SCX") ) then
          ra_scx  = real64Attribute(aSet,"RA_SCX") *degreesToRadians
          dec_scx = real64Attribute(aSet,"DEC_SCX")*degreesToRadians
          pa_scx  = real64Attribute(aSet,"PA")     *degreesToRadians
          rollAngle = pa_scx
      end if
      omFilter = stringAttribute(aSet, "FILTER")
      dataMode = stringAttribute(aSet, "DATAMODE") !needed by cal

      if (trim(dataMode) .eq. "IMAGING") then 
         OM_MODE = IMAGE
      else
         OM_MODE = FAST
      end if

      !************************************
      ! Get the filter Id integer value
      !************************************
      filterIdInt = filterIdToInt(omFilter)

      !*********************
      ! Set the CAL state
      !*********************
      call CAL_setState(OM, ccdModeId=OM_MODE, filterId = filterIdInt)

      !***********************************************************
      ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
      !***********************************************************
      if( hasBlock( aSet, "MODES") ) then
          call readModes(aSet, x0, y0, xsize, ysize)  
      else
          x0  = int32Attribute(aSet, "WINDOWX0")
          y0  = int32Attribute(aSet, "WINDOWY0")
          xSize  = int32Attribute(aSet, "WINDOWDX")
          ySize  = int32Attribute(aSet, "WINDOWDY")
      end if



      call CAL_omGetPlateScale(lambda, plateScale) ! (arcsecs per pixel) 
      plateScale = plateScale * degreesToRadians / 3600.0_double !  to radians

      !***********************************************************
      ! Get the number of seconds since the mission reference time
      !***********************************************************
      time =  MissionTime(aSet)
      call CAL_GetBoresightMatrix(time, boremat) ! boresight offset matrix

      !*******************************************************************
      ! need inverted boresight matrix which is obtained via the transpose. 
      !********************************************************************
      inv_boremat = TRANSPOSE(boremat)
      attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle)

      !*******************************************************
      ! This routine requires a matrix of size 0:2, 0:2, but
      ! after that we want it to be of size 1:3, 1:3
      !******************************************************
      call OAL_toAttitudeMatrix(attitude, attmatrix1)
      attmatrix=attmatrix1

      !**********************************************
      ! rotationCorrection corrects for OM axes definition.
      !**********************************************
      rollAngle = rollAngle + rotationCorrection

      !************************************************************************
      ! converts RA and DEC from spherical cordinates to cartesian coordinates
      !************************************************************************
      call sla_dcs2c(ra_scx, dec_scx, pointvec)

      !********************************************************
      ! normalize the vector pointvec and return in pointvec2 -
      ! r1 holds the mdulus of the vector. pointvec2 will store
      ! the unit-vector in the direction of pointvec
      !********************************************************
      call sla_dvn(pointvec, pointvec2, r1)

      pointvec(1) = -pointvec2(1)
      pointvec(2) = -pointvec2(2)
      pointvec(3) =  pointvec2(3)
      !------------------------------------------------------------------------

      euler_axis = -rollangle * pointvec       

      ! convert euler axis vector to a rotation matrix
      call sla_dav2m(euler_axis, rotmat) 

      ! initialise output vector
      omvec = 0.0_double

      !************************************************************************
      !  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
      !  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
      !  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
      !
      !  This creates omvec, the om boresight vector measured in the ERF
      !  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
      !  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
      !  is the OM boresight as measured in the OM reference frame
      !************************************************************************
      att_invbor = matmul(attmatrix, inv_boremat)
      omvec = matmul(att_invbor, ompoint)

      !************************************************************************
      ! converts the omvec cartesian vector to ra and dec spherical coordinates
      ! and check RA lies within sensible range
      ! ra_point and ra_dec refer to the pointing direction of the instrument
      !************************************************************************
      call sla_dcc2s(omvec, ra_point, dec_point)
      if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi

      write(errstr, '(a, f8.4)') "RA of OM boresight: ", &
           ra_point * radiansToDegrees
      call message(AppMsg, Verbose)
      write(errstr, '(a, f8.4)') "DEC of OM boresight: ", &
           dec_point * radiansToDegrees
      call message(AppMsg, Verbose)


    !**********************************************************************
    !
    !
    !**********************************************************************

    if( hasAttribute(aset, "CDELT1") ) then
        cdelt1 = real64Attribute( aSet, "CDELT1")  ! RA plate-scale (degrees and negative)
        cdelt2 = real64Attribute( aSet, "CDELT2")  ! Dec plate-scale (degrees)
        crpix1 = real64attribute( aSet, "CRPIX1")
        crpix2 = real64attribute( aSet, "CRPIX2")
        crval1 = real64attribute( aSet, "CRVAL1")
        crval2 = real64attribute( aSet, "CRVAL2")
        isSkyImage = .true.
    else
        isSkyImage = .false.
 
    end if    
 
end subroutine InitialiseCelestial

!*******************************************
! Function MissionTime - Returns the number
! of seconds from the start of the mission
! to the observation time. Added by CB because
! the call to CAL_GetBoresiteMatrix needs
! the time in seconds from the launch.
!******************************************
double precision function  MissionTime(primaryTable)
    type(DataSetT), intent(in)   :: primaryTable

    character(19)  :: dateObs         
    character(19)  :: dateEnd        
    real(double)  :: JulianDate0, JulianDate1, JulianDate2
    character(19)  :: missionReferenceTime = "1998-01-01T00:00:00"

    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the start of the observations.
    dateObs = stringAttribute(primaryTable,  "DATE-OBS")    
    JulianDate1 = ModifiedJulianDate(dateObs)

    ! Compute the Julian date of the end of the observations.
    dateEnd = stringAttribute(primaryTable,  "DATE-END")   
    JulianDate2 = ModifiedJulianDate(dateEnd)

    ! The required time is the difference between the average of the 
    ! start and end times and the mission reference time, in seconds.
    MissionTime = ((JulianDate1 + JulianDate2) / 2.0d0 - JulianDate0) * 86400.0d0
end function MissionTime



!*******************************************************
! ModifiedJulianDate - Returns the modified Julian date
! from the string date representation
!*******************************************************
real(double) function ModifiedJulianDate(dateString)
      character(*), intent(in) :: dateString

    integer :: year            ! Stores the year from the date string
    integer :: month           ! Stores the month from the date string
    integer :: day             ! Stores the day from the date string  
    integer :: hour            ! Stores the hour from the date string
    integer :: minute          ! Stores the minute from the date string  
    integer :: second          ! Stores the second from the date string
    character(1) char    ! Used for date and time extraction 
    ! from string representation
    double precision JulianDay    
    integer :: flag

    read(dateString, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
           year, char, month, char, day, char, hour, &
           char, minute, char, second     

    ! Get the modified Julian date for this date
    call sla_CLDJ(year, month, day, JulianDay, flag)

    ! Need to add on the fraction of a day corresponding to the time
    ModifiedJulianDate = JulianDay + (dble(hour) + &
           (dble(minute) + dble(second) / 60.0d0) / 60.0d0) / 24.0d0
end function ModifiedJulianDate


!*****************************************************************
!
!
!
!****************************************************************
subroutine computeRaDec(source ) 
  
    type(sourceParameters), intent(inout):: source
    
    real(double) :: dx, dy
  
   
    dx = (dble(source%xpos) - crpix1) * cdelt1
    dy = (dble(source%ypos) - crpix2) * cdelt2
    call convertxytoradec(dx, dy, crval1, crval2, source%rightAscension, source%declination )


  
 end subroutine computeRADEC

!********************************************************************
!
!
!
!
!********************************************************************
subroutine convertxytoradec(x, y, ra0, dec0, ra, dec)

    real(double), intent(in) :: x, y, ra0, dec0
    real(double), intent(out) :: ra, dec


    call sla_dtp2s(x * degreesToRadians, y * degreesToRadians,  &
               ra0 * degreesToRadians, dec0 * degreesToRadians, &
               ra, dec)

    ra = ra * radiansToDegrees
    dec = dec * radiansToDegrees


end subroutine convertxytoradec



end module omsourcecelestial_mod


